#!/usr/bin/env python

# Rnmap console client v. 0.10
#
# Copyright (C) 2000-2003 Tuomo Makinen (tmakinen@pp.htv.fi)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Rnmap cli client v. 0.10

Copyright (C) 2000-2003 Tuomo Makinen
Redistributable under the terms of the GNU General Public License

Usage:
    -s address       Rnmap server address
    -p port          Server port (default 3418)
    -f file          Filename to save the scanning result (no stdout)
    -n cmd           Nmap command
    -u user          Username
    -c file          Text file containing Username, Password and Server:Port information
                     (Newline separated)
                     
    --help           Prints this message
    --nossl          Turns SSL support off
    
    --mparseable     Use machine parseable log form
    --scriptkiddie   Use script kiddie log form
    --xml            Use xml log form
    
    --version        Show version of rnmap cli client

    example:
    # rnmap.py -s 192.168.1.78 -n "-sS -p 1-65535 192.168.1.10" --mparseable"""

import sys
import os
import string
import getopt
import signal
import getpass
import stat
import signal
    
sys.path[:0] = ['../lib']

import clientsocket
from client import *

Port = 3418
Server = None
Mode = " -oN"
Nmapcmd = ""
Output_file = None
Conf_file = None

# Set default umask
os.umask(0066)

def open_connection(message):
    con_stat = Connection.open_connection(Server, Port)
    if con_stat[0] == -1:
        handle_error(con_stat[1])
    if Connection.write_socket(message) == -1:
        handle_error(SENDERR)
    workbuf = ''; cnt = 0
    while 1:
        chunk = Connection.read_socket()
        if not chunk:
            break
        elif chunk == -1:
            handle_error(RECVERR)
        workbuf = workbuf + chunk
        while 1:
            index = string.find(workbuf, CRLF)
            if index == -1:
                break
            handle_message(workbuf[:index], cnt)
            workbuf = workbuf[index+2:]
            cnt = cnt + 1

    # Close connection
    Connection.close_socket()

def handle_error(errmsg):
    Connection.close_socket()
    fatalerror(errmsg + "\n")

def verify_cb(conn, cert, errnum, depth, ok):
    if not ok:
        print "\n** Can't verify certificate ** Error " + `errnum` + " at depth " + `depth`\
              + "\n\n Issuer = %s\n Subject = %s\n MD5 Digest = %s\n" % \
              (cert.get_issuer(), cert.get_subject(), digest_cert(crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)))
        if errnum == 18:    # Self signed certificate
            try:
                valid = raw_input("Do you accept certificate [y/n]? ")
            except KeyboardInterrupt:
                ok = 0
            if valid == 'y' or valid == 'Y':
                if store_certificate(crypto.dump_certificate(crypto.FILETYPE_PEM, cert)) == -1:
                    sys.stderr.write("Can't save certificate.\n")
                sys.stdout.write("\n"); ok = 1
            else:
                ok = 0
        if not ok:
            print "\nCertificate was rejected.\n"
    return ok

def read_configuration():
    try:
        fp = open(Conf_file).readlines()
    except IOError, e:
        fatalerror("Can't open " + Conf_file + ": " + e[1] + "\n")
    if stat.S_IMODE(os.lstat(Conf_file)[stat.ST_MODE]) & 077:
        fatalerror("File " + Conf_file + " must have no more than -rwx------ (700) permissions.\n")
    else:
        c = len(fp) - 1
        while c >= 0:
            if fp[c] == "\n":
                del fp[c]
            c = c - 1
        try:
            logid = fp[0][:-1]
            passwd = fp[1][:-1]
            pindex = string.rfind(fp[2], ":")
            server = fp[2][:pindex]
            port = string.atoi(fp[2][pindex + 1:-1])
        except:
            fatalerror("Error reading configuration file!\n")
    return server, port, logid, passwd

def ask_username():
    try:
        logid = raw_input("Username for " + Server + ": ")
    except KeyboardInterrupt:
        fatalerror("\n")
    return logid

def ask_password():
    try:
        passwd = getpass.getpass()
    except KeyboardInterrupt:
        fatalerror("\n")
    return passwd

def handle_message(server_msg, round):
    if round == 0:
        if server_msg[:3] != "201":
            display_exception(server_msg[:3])
            Connection.close_socket()
            sys.exit()
        server_msg = server_msg[4:]
    output_message(server_msg)
    
def output_message(server_msg):
    if Output_file == None:
        sys.stdout.write(server_msg)
    else:
        try:
            Scan_result.write(server_msg)
            Scan_result.flush()
        except IOError, e:
            sys.stderr.write("Can't write to disk: " + e[1] + "\n")

def display_exception(code):
    try:
        sys.stdout.write(PROTOCOL_CODES[code] + "\n")
    except KeyError:
        sys.stderr.write(PROTOCOL_ERROR + "\n")
        
def signal_handler_term(signum, frame):
    Connection.write_socket(INTERRUPT_SCAN)

def fatalerror(error_message):
    sys.stderr.write(error_message)
    sys.exit(1)

# --------------------------------------------------------------------
try:
    opts, args = getopt.getopt(sys.argv[1:], 's:p:n:f:u:c:',\
    ['help', 'version', 'mparseable', 'scriptkiddie', 'xml', 'nossl'])
except:
    fatalerror("Try 'rnmap.py --help' for more information.\n")
for o, a in opts:
    if o == '--version':
        sys.stdout.write("Rnmap cli client v. 0.10\n")
        sys.exit()
    if o == '-s':
        Server = a
    if o == '-p':
        try:
            Port = string.atoi(a)
        except ValueError:
            fatalerror("Invalid port specification.\n")
    if o == '-n':
        Nmapcmd = string.replace(a, ":", "###")
    if o == '-f':
        Output_file = a
    if o == '-u':
        Logid = a
    if o == '-c':
        Conf_file = a
    if o == '--mparseable':
        Mode = " -oM"
    elif o == '--scriptkiddie':
        Mode = " -oS"
    elif o == '--xml':
        Mode = " -oX"
    if o == '--nossl':
        clientsocket.UseSSL = 0
    if o == '--help':
        print __doc__
        sys.exit()

if clientsocket.UseSSL == -1:
    sys.stderr.write("Can't find pyOpenSSL module. If you are connecting to non SSL Rnmap server use '--nossl' command line option.\n")
    sys.exit(1)
    
if Conf_file == None:
    if Server == None:
        print __doc__
        sys.exit(1)
    try:
        if Logid:
            Passwd = ask_password()
    except NameError:
        Logid = ask_username()
        Passwd = ask_password()
else:
    Server, Port, Logid, Passwd = read_configuration()

if Output_file:
    try:
        Scan_result = open(Output_file, "w")
    except IOError, e:
        fatalerror("Can't write to disk: " + e[1] + "\n")

# --------------------------------------------------------------------
# Set signal handlers
signal.signal(signal.SIGALRM, signal_handler_sigalrm)
signal.signal(signal.SIGTERM, signal_handler_term)
signal.signal(signal.SIGINT, signal_handler_term)

# Client object
if clientsocket.UseSSL:
    Connection = clientsocket.SSLClient(verify_cb, CA_CERTS)
    crypto = clientsocket.crypto
else:
    Connection = clientsocket.RegularClient()
    
# Open connection
open_connection("101 " + Logid + ":" + Passwd + ":" + Nmapcmd+Mode)

if Output_file:
    Scan_result.close()
    
sys.exit()
