#!/usr/bin/env python

# Rnmap server v. 0.10
#
# Copyright (C) 2000-2003 Tuomo Makinen (tmakinen@pp.htv.fi)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Rnmap server v. 0.10

Copyright (C) 2000-2003 Tuomo Makinen
Redistributable under the terms of the GNU General Public License

Usage:
    -p port         Listen on specified port (default: 3418)
    -n path         Nmap path (default: /usr/bin/nmap and /usr/local/bin/nmap)
    -h address      Listen address (default: 0.0.0.0)
    -r dir          Directory to chroot to after initialization
    -u uid          Run rnmap server as a specified uid after initialization
    -g gid          Run rnmap server as a specified gid after initialization
    -c dir          Configuration files directory (default .)
    -l dir          Log file directory (default is same as configuration files)
    
    --help          Prints this message
    --version       Show version of rnmap server
    
    --nofilter      Turns filtering rules off
    --nossl         Turns SSL support off
"""

import os
import string
import sys
import md5
import time
import getopt
import signal
import base64
import select
import fcntl
from errno import EINTR

if sys.hexversion < 0x20200f0:
    import FCNTL
else:
    FCNTL = fcntl
    
sys.path[:0] = ['../lib']

import serversocket

## Globals
#----------------------------------------------------------
# Protocol message type codes
SCANREQUEST   = "101"
INTERRUPTSCAN = "102"
SCANRSLT      = "201"
AUTHDENI      = "401"
SCANFAIL      = "402"
PROTOERROR    = "403"
BADCHARS      = "404"

# NOTHESECHARS is tuple which has characters that can cause
# unwanted action, and we don't want pass them to system.
NOTHESECHARS = ("|",
                ";",
                "<",
                ">",
                "&")

Server = ''
Port = 3418
Nmapl = "/usr/bin/nmap"
Filteruse = 1
Chroot_stat = 0
Process_uid = None
Process_gid = None
Active_childs = []

try:
    Conf_dir = os.environ["RNMAPDIR"]
except KeyError:
    Conf_dir = os.getcwd()
#----------------------------------------------------------

#==========================================================================
# read configuration files ...
#==========================================================================
# filter.conf is optional file, if omitted we're not using filtering rules.

def confreader():
    global Filteruse, Userlist, Clist, Fmethod
    Userlist = {}
    if Filteruse == 1:
        try:
            fp = open(Fconf_file)
        except IOError:
            Filteruse = None
            Fmethod = None
        else:
            tmplist = fp.readlines()
            Clist = []
            for i in xrange(len(tmplist)):
                if "#" in tmplist[i]:
                    continue
                elif tmplist[i][0:] == "\012":
                    continue
                elif tmplist[i][0] == "!":
                    Fmethod = tmplist[i][9:16]
                else:
                    Clist.append(tmplist[i])
            for j in xrange(len(Clist)):
                Clist[j] = Clist[j][:-1]
            fp.close()
    try:
        fp = open(Ulist_file)
    except IOError:
        return -1
    else:
        ul = fp.readlines()
        fp.close()
        for i in xrange(len(ul)):
            stopsign = 0
            for j in xrange(len(ul[i])): 
                if ul[i][0] == "#":
                    continue
                elif ul[i][0] == " ":
                    continue
                elif ul[i][j] == ":" and j != 1:
                    if stopsign == 0:
                        Userlist[ul[i][:j]] = base64.decodestring(ul[i][j+1:])
                        stopsign = 1

#==========================================================================

def wait_connection():
    # Activate server
    Servobj.do_socket()
    serv_stat = Servobj.activate_server(Server, Port)
    if serv_stat[0] == -1:
        fatalerror(serv_stat[1])
    # Log server startup
    log_startup()
    # Serve forever
    while 1:
        try:
            rlist = select.select([Servobj.sockobj], [], [], 5.0)[0]
        except select.error, e:
            if e[0] == EINTR:
                continue
        else:
            if Servobj.sockobj in rlist:
                try:
                    conn, addr = Servobj.accept_conn()
                except:
                    pass
                else:
                    reap_childrens()
                    try:
                        pid = os.fork()
                    except OSError:
                        conn.close()
                        continue
                    if pid == 0:
                        # Close listening socket
                        Servobj.close_socket()
                        # Set default signal handling for child
                        signal.signal(signal.SIGTERM, signal.SIG_DFL)
                        signal.signal(signal.SIGHUP, signal.SIG_DFL)
                        # --
                        handle_connection(conn, addr[0])
                        os._exit(0)
                    else:
                        Active_childs.append(pid)
                        # Close connected socket
                        conn.close()
            else:
                reap_childrens()

def reap_childrens():
    while Active_childs:
        try:
            pid, status = os.waitpid(0, os.WNOHANG)
        except OSError:
            pid = None
        if not pid:
            break
        Active_childs.remove(pid)
                
def handle_connection(conn, host):
    if Filteruse != None and filter(host) == 1:
        logwriter("banned_host", host)
        Servobj.write_socket(conn, AUTHDENI)
    else:
        client_msg = Servobj.read_socket(conn)
        if not client_msg:  # Connection is probably broken
            pass
        elif client_msg[:3] == SCANREQUEST:
            process_message(client_msg[4:], host, conn)
        else:
            Servobj.write_socket(conn, PROTOERROR)
    if serversocket.UseSSL:
        conn.shutdown()
    conn.close()
    
def process_message(client_msg, host, conn):
    usr_point = string.find(client_msg, ":")
    cmd_point = string.rfind(client_msg, ":")
    if usr_point != cmd_point:
        login = client_msg[:usr_point]
        cmd = client_msg[cmd_point+1:]
        passwd = md5.new(client_msg[usr_point+1:cmd_point]).digest()
        cmd = string.replace(cmd, "###", ":")
        auth_stat = authenticate(login, passwd)
        if auth_stat == 1:
            scanstat = 1
            for i in xrange(len(cmd)): # Check chars 
                if cmd[i] in NOTHESECHARS:
                    scanstat = 0; break
            if scanstat == 0:
                logwriter("illegal_parameters", host, login, cmd)
                Servobj.write_socket(conn, BADCHARS)
            elif scanstat == 1:
                start_scanner(login, host, cmd, conn)
        elif auth_stat == 0:
            logwriter("bad_password", host, login)
            Servobj.write_socket(conn, AUTHDENI)
        elif auth_stat == None:
            logwriter("unknown_user", host, login)
            Servobj.write_socket(conn, AUTHDENI)
    else:
        Servobj.write_socket(conn, PROTOERROR) 
    
def start_scanner(login, host, cmd, conn):
    outchunk = ''
    client_msg = ''
    connected = 1
    chunk_counter = 0
    args = [Nmapl,] + string.split(cmd + ' - ')
    # --
    pRc, cWp = os.pipe()
    try:
        child_pid = os.fork()
    except OSError:
        os._exit(1)
    if child_pid == 0:
        os.dup2(cWp, 1)     # Combine stderr and stdout
        os.dup2(cWp, 2)     ###########################
        # Ensure that stray fd's are closed
        for i in xrange(3, 256):
            try:
                os.close(i)
            except:
                pass
        try:
            os.execvp(Nmapl, args)
        except:
            os._exit(1)
    os.close(cWp)
    fromchild = os.fdopen(pRc, "r")
    # --
    fcntl.fcntl(fromchild.fileno(), FCNTL.F_SETFL, os.O_NONBLOCK)
    outfd = fromchild.fileno()
    connfd = conn.fileno()
    # --
    while 1:
        rlist = select.select([outfd, connfd], [], [])[0]
        if outfd in rlist:
            outchunk = fromchild.read()
            if not outchunk:
                break
            else:
                if chunk_counter == 0:
                    outchunk = SCANRSLT + " " + outchunk
                Servobj.write_socket(conn, outchunk)
            chunk_counter = chunk_counter + 1   
        if connfd in rlist:
            client_msg = Servobj.read_socket(conn)
            if not client_msg:
                connected = 0
            if not connected or client_msg == INTERRUPTSCAN:
                # Send SIGTERM to process
                try:
                    os.kill(child_pid, signal.SIGTERM)
                except OSError:
                    pass
        time.sleep(0.1)
    # --
    # Close pRc and reap children
    os.close(pRc)
    try:
        pid, status = os.waitpid(child_pid, 0)
    except OSError:
        pass
    # Write log
    if not connected:
        logwriter("closed_connection", host)
    elif client_msg == INTERRUPTSCAN:
        logwriter("interrupted_scan", host, login, cmd)
    else:
        logwriter("succesful_scan", host, login, cmd)

def authenticate(login, passwd):
    if Userlist.has_key(login):
        if Userlist[login] == passwd:
            return 1
        else:
            return 0

def filter(host):
    if Fmethod == "drpexcp":
        if host in Clist:
            return 0
        else:
            return 1
    elif Fmethod == "allexcp":
        if host in Clist:
            return 1
        else:
            return 0

def logwriter(information, param1=None, param2=None, cmd=None):
    pid = os.getpid()
    if information == "succesful_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s performed scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information ==  "bad_password":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s tried to connect with wrong password.\n"\
               % (pid, param2, param1)
    elif information == "unknown_user":
        wlog = "[" + time.ctime(time.time()) + "][%d] illegal user %s from %s tried to connect.\n"\
               % (pid, param2, param1)
    elif information == "failed_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s performed failed scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "banned_host":
        wlog = "[" + time.ctime(time.time()) + "][%d] banned host: %s dropped.\n"\
               % (pid, param1)
    elif information ==  "illegal_parameters":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s used illegal parameters: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "interrupted_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s interrupted scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "closed_connection":
        wlog = "[" + time.ctime(time.time()) + "][%d] remote host %s closed connection prematurely.\n"\
               % (pid, param1)
    elif information == "startup":
        if serversocket.UseSSL:
            wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server started with SSL support.\n"\
                   % (param1)
        else:
            wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server started without SSL support.\n"\
                   % (param1)
        wlog_listen = "[" + time.ctime(time.time()) + "][%d] rnmap server listening on port %d\n"\
                % (pid, param2)
        wlog = wlog + wlog_listen
    elif information == "sigterm":
        wlog = "[" + time.ctime(time.time()) + "][%d] received the SIGTERM signal.\n"\
               % (pid)
    elif information == "sighup":
        wlog = "[" + time.ctime(time.time()) + "][%d] received the SIGHUP signal.\n"\
               % (pid)
    elif information == "ulist_failed":
        wlog = "[" + time.ctime(time.time()) + "][%d] error reading users.list.\n"\
               % (pid)
    elif information == "chrooted":
        wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server chrooted to %s\n" \
               % (pid, Chrootdir)
    elif information == "failed_startup":
        wlog = "[" + time.ctime(time.time()) + "][%d] %s" % (pid, param1)
    #---
    fp = open(Rnmapdlog, "a")
    fp.write(wlog)
    fp.close()

def signal_handler_sighup(signum, frame):
    logwriter("sighup")
    if confreader() == -1:
        logwriter("ulist_failed")

def signal_handler_sigterm(signum, frame):
    logwriter("sigterm")
    try:
        os.unlink(Rnmapdpid)
    finally:
        sys.exit()

def fatalerror(error):
    logwriter("failed_startup", error + " - server startup failed.\n")
    sys.exit(1)
    
class nulldevice:
    def write(self, s):
        pass
    
def chroot_process():
    try:
        os.chroot(Chrootdir)
    except AttributeError:
        fatalerror("can't chroot() (you need python 2.2 or newer for it)")
    except OSError, e:
        fatalerror("can't chroot() to " + e.filename + ": " + e.strerror)
        
def become_daemon():
    try:
        pid = os.fork()
    except (OSError, AttributeError):
        fatalerror("Can't fork()")
    if pid != 0:
        sys.exit()
    else:
        os.setsid()
        if Chroot_stat == 1:
            chroot_process()
        os.chdir('/')
        os.umask(0066)
        sys.stdin.close()
        sys.stdout = nulldevice()
        sys.stderr = nulldevice()
        # Ensure that stray fd's are closed
        for i in range(3, 256):
            try:
                os.close(i)
            except:
                pass

def log_startup():
    # Write pid to disk
    try:
        fp = open(Rnmapdpid, "w")
    except IOError:
        fatalerror("can't write pid file to disk")
    my_pid = os.getpid()
    fp.write(`my_pid` + "\012")
    fp.close()
    logwriter("startup", Pid, Port)
    if Chroot_stat == 1:
        logwriter("chrooted")
    if Conf_stat == -1:
        logwriter("ulist_failed")
    
        
# Fetch command line parameters.
try:
    opts, args = getopt.getopt(sys.argv[1:], 'p:n:h:r:c:l:u:g:',\
    ['help', 'nofilter', 'nossl', 'version'])
except:
    sys.stderr.write("Try 'rnmap_server.py --help' for more information.\n")
    sys.exit(1)

for o, a in opts:
    if o == '-h':
        Server = a
    if o == '-p':
        Port = a
    if o == '-n':
        Nmapl = a
    if o == '-r':
        Chroot_stat = 1
        Chrootdir = a
    if o == '-u':
        Process_uid = a
    if o == '-g':
        Process_gid = a
    if o == '-c':
        Conf_dir = a
    if o == '-l':
        Log_dir = a
    if o == '--help':
        print __doc__
        sys.exit()
    if o == '--version':
        print "Rnmap server v. 0.10"
        sys.exit()
    if o == '--nofilter':
        Filteruse = None
        Fmethod = None
    if o == '--nossl':
        serversocket.UseSSL = 0

if Chroot_stat == 1:
    if Conf_dir == os.getcwd():
        Conf_dir = './'
try:
    if Log_dir:
        pass
except NameError:
    Log_dir = Conf_dir

# Check SSL support
if serversocket.UseSSL == -1:
    sys.stderr.write("Can't find pyOpenSSL module. If you want to start " +\
        "non SSL Rnmap server use '--nossl' command line option.\n")
    sys.exit(1)
    
# Few globals
Fconf_file = os.path.join(Conf_dir, "filter.conf")
Ulist_file = os.path.join(Conf_dir, "users.list")
Rnmapdpid = os.path.join(Conf_dir, "rnmapd.pid")
Rnmapdlog = os.path.join(Log_dir, "rnmap_server.log")
if serversocket.UseSSL:
    Pkey_file = os.path.join(Conf_dir, "rnmapd.key")
    Cert_file = os.path.join(Conf_dir, "rnmapd.crt")

# Create server object
if serversocket.UseSSL:
    if os.access(Pkey_file, os.R_OK) == 0:
        sys.stderr.write(Pkey_file + " doesn't exists! You must generate private key and"\
               + " certificate for SSL server.\n")
        sys.exit(1)
    if os.access(Cert_file, os.R_OK) == 0:
        sys.stderr.write(Cert_file + " doesn't exists! You must generate private key and"\
               + " certificate for SSL server.\n")
        sys.exit(1)
    Servobj = serversocket.SSLServer(Pkey_file, Cert_file)
else:
    Servobj = serversocket.RegularServer()

# Get pid
Pid = os.getpid()

# Read configuration files
Conf_stat = confreader()

# Daemonize
become_daemon()

# Set signal handlers
signal.signal(signal.SIGTERM, signal_handler_sigterm)
signal.signal(signal.SIGHUP, signal_handler_sighup)

# setgid/setuid
if Process_gid != None:
    try:
        os.setgid(string.atoi(Process_gid))
    except:
        fatalerror("invalid gid specification")
        
if Process_uid != None:
    try:
        os.setuid(string.atoi(Process_uid))
    except:
        fatalerror("invalid uid specification")

# Check -p option
if type(Port) == type(''):
    try:
        Port = string.atoi(Port)
    except ValueError:
        fatalerror("invalid port specification")
        
# Check if nmap exists
if os.access(Nmapl, os.F_OK) == 0:
    Nmapl = "/usr/local/bin/nmap"
    if os.access(Nmapl, os.F_OK) == 0:
        fatalerror(Nmapl + " doesn't exists! Specify Nmap path with -n option")

# Start work
wait_connection()
